/*
 * Png stitcher.
 *
 * Png stitcher is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Png stitcher is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2007 Andrei Krger */

// #define deb(fmt, args...) printf("%s: " fmt, __FUNCTION__, ##args)
#define deb(fmt, args...) while(0) {}

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <png.h>
#include <setjmp.h>
#include "aspnger.h"

#define C(r,g,b) ((r<<16) | (g<<8) | b)
#define red(c) ((c>>16)&0xff)
#define green(c) ((c>>8)&0xff)
#define blue(c) ((c)&0xff)

/* Poistaa Karttapaikan "peruskartan" korkeuskyrien ruskean, aluperisen (0) ja varsinaiset (1).
 * Keltaisen (2) ja sinisen (3) pllevrjtyt alueet palautetaan alkuperisiksi. */
void clean_karttapaikka8000(struct image *img)
{
  long x, y;
  u8 **data = img->data;

  for(y = 0; y < img->height; y++)
  {
    for(x = 0; x < 3*img->width; x+=3)
    {
      int col = (data[y][x]<<16) | (data[y][x+1]<<8) | data[y][x+2];
      if (col == C(171,87,0) ||
          col == C(192,129,63) || col == C(213,171,127) || col == C(234,213,191))
      { /* puhdas ruskea ja sen vaaleammat svyt valkoiseksi */
        data[y][x] = data[y][x+1] = data[y][x+2] = 255;
      }
      else if (col == C(192,129,38) || col == C(213,171,76) || col == C(234,213,114))
      { /* ruskean ja keltaisen sekoitus keltaiseksi */
        data[y][x] = data[y][x+1] = 255; data[y][x+2] = 153;
      }
      else if (col == C(157,129,63) || col == C(143,171,127) || col == C(129,213,191))
      { /* ruskean ja sinisen sekoitus siniseksi */
        data[y][x] = 115; data[y][x+1] = data[y][x+2] = 255;
      }
    }
  }
}

#define ARRAY_SIZE(a) (sizeof(a)/(sizeof((a)[0])))

/* Poistaa Karttapaikan 1:40000 korkeyskyrien ruskean ja avokallion harmaan. *
 * 1:40000 perustuu 2x2 antialiasointiin. */

struct replace { u32 from; u32 to; };
struct replace_chain { long count; struct replace *chain; };

static struct replace_chain k40_hash[256];

#define white 0xffffff
#define nop(r,g,b) {C(r,g,b), C(r,g,b)}
#define DEL(r,g,b) {C(r,g,b), white}

struct replace k40000_colors[] = {
  /* vesistt */
  nop(128,255,255),       /*  0: sininen, vesi */
  DEL(199,235,235),       /*  0: syanni, suo */
  DEL(230,245,245),       /*  0: syanni2, suo2 */
  
  /* alueet */
  nop(230,230,128),       /* 10: vihre, kentt-1, tms.*/
  nop(230,255,128),       /* 10: vihre, puisto? */
  nop(179,255,128),       /* 10: vihre, tarha? */
  nop(230,230,102),       /* 10: vihre, teollisuusalue! */
  nop(179,217, 0),        /* 10: vihre, vino viivoitus??? */
  nop(235,235,235),       /* 10: harmaa, varastoalue */
  DEL(209,199,209),       /* 10: harmaa, avokallio */
  nop(255,255,64),        /* 10: keltainen, aukea? */
  nop(255,217,128),       /* 10: keltainen, pelto? */
  nop(209,204,64),        /* 10: vihre, tiheikk? */
  nop(224,194,224),       /* 10: violetti, asuinalue? */
  nop(204,153,204),       /* 10: violetti, teollisuus? */
  nop(255,128,102),       /* 10: oranssi, teollisuus? */
  
  /* korkeuskyrt */
  DEL(166, 77, 0),        /* 20: ruskea, korkeyskyrt */
  
  /* rauhoitusalueet */
  nop(0,230,0),           /* 30: vihre, rauhoitusalue */
  
  /* ojat */
  nop(0,128,255),         /* 40: sininen, ojat/teksti */
  
  /* rajat, tienumerot */
  nop(0,0,255),           /* 45: sininen, kuntaraja? */
  nop(230,0,255),         /* 45: violetti*/
  
  /* tekstit, tiet */
  nop(128, 64, 0),        /* 50: ruskea, teksti */
  nop(0,179,0),           /* 50: vihre, teksti */
  nop(0,51,255),          /* 50: sininen, teksti */
  nop(204,0,0),           /* 50: punainen */
  
  /* tiet, polut */
  nop(0, 0, 0),           /* 90: musta*/
  
};

#define mix4(a,b,c,d) ((a+b+c+d)/4)
#define MIX(c4) C( \
    mix4(  red(c4[0]),  red(c4[1]),  red(c4[2]),  red(c4[3])), \
    mix4(green(c4[0]),green(c4[1]),green(c4[2]),green(c4[3])), \
    mix4( blue(c4[0]), blue(c4[1]), blue(c4[2]), blue(c4[3])))

void k40_add_transfer(u32 from, u32 to)
{
  u8 hash = (u8)(from >> 16) ^ (u8)(from >> 8) ^ (u8)(from);
  struct replace_chain *chain = &k40_hash[hash];
  int i;

  /* Check if transfer already exists */
  for(i=0; i < chain->count; i++)
  {
    if (chain->chain[i].from == from) {
      /*
      if (chain->chain[i].to != to)
        printf("confligt: %06lx --> %06lx or %06lx\n", from, chain->chain[i].to, to);
      */
      return;
    }
  }

  chain->chain = realloc(chain->chain, (chain->count+1)*sizeof(struct replace));
  chain->chain[chain->count].from = from;
  chain->chain[chain->count].to = to;
  chain->count++;
}

void loop_colors(int filled, u32 *src, u32 *dest, int start)
{
  int i;
  int a;

  if (filled == 4) {
    k40_add_transfer(MIX(src), MIX(dest));
    return;
  }

  while (filled < 4) 
  {
    for (i=filled; i<4; i++) {
      src[i] = white;
      dest[i] = white;
    }
    k40_add_transfer(MIX(src), MIX(dest));

    for (a=start; a < ARRAY_SIZE(k40000_colors); a++)
    {
      src[filled] = k40000_colors[a].from;
      dest[filled] = k40000_colors[a].to;
      loop_colors(filled+1, src, dest, a);
    }
    filled++;
  }
}

void init_k40000() {

  int i,j,a;
  u32 src[4];
  u32 dest[4];
  
  /* Optimized: add plain colors (4 different brightness) to the beginning of hash chains. */
  for(i = 4; i ; i--) {
    for(a = 0; a < ARRAY_SIZE(k40000_colors); a++)
    {
      for(j = 0; j < 4; j++) {
        src[j] = (j < i) ? k40000_colors[a].from : white;
        dest[j] = (j < i) ? k40000_colors[a].to : white;
      }
      k40_add_transfer(MIX(src), MIX(dest));
    }
  }

  /* Find transfered colors and loop all variants */
  for (a=0; a < ARRAY_SIZE(k40000_colors); a++)
  {
    if (k40000_colors[a].from != k40000_colors[a].to)
    {
      for(i = 3; i ; i--) {
        for(j=0; j < i; j++) {
          src[j] = k40000_colors[a].from;
          dest[j] = k40000_colors[a].to;
        }
        loop_colors(i, src, dest, 0);
      }
    }
  }
}

void clean_karttapaikka40000(struct image *img)
{
  long x, y;
  u8 **data = img->data;

  for(y = 0; y < img->height; y++)
  {
    for(x = 0; x < 3*img->width; x+=3)
    {
      u32 col = (data[y][x]<<16) | (data[y][x+1]<<8) | (data[y][x+2]);
      u8 hash = (data[y][x]) ^ (data[y][x+1]) ^ (data[y][x+2]);
      long i;
      struct replace *rpl;

      /* white */
      if (col == 0xffffff)
        continue;

      for(i = 0, rpl = k40_hash[hash].chain; i < k40_hash[hash].count; i++, rpl++) {
        if (col == rpl->from) {
          data[y][x]   = (u8)(rpl->to >> 16);
          data[y][x+1] = (u8)(rpl->to >> 8);
          data[y][x+2] = (u8)(rpl->to);
          break;
        }
      }
    }
  }
}

/* Vrit AS-ruskealla */

/* PMS 138C (?) */
#define PMS138C_R 204
#define PMS138C_G 102
#define PMS138C_B 0

void colorize_brown(struct image *img)
{
  long x, y;
  long r,g,b,Y;
  long R[256],G[256],B[256]; /* fast color lookups */
  long density = 90; /* 10% of original color, 90% of brown */
  u8 **data = img->data;

  for(Y=0;Y<256;Y++) {
    R[Y] = PMS138C_R + (Y*(255-PMS138C_R)/255);
    G[Y] = PMS138C_G + (Y*(255-PMS138C_G)/255);
    B[Y] = PMS138C_B + (Y*(255-PMS138C_B)/255);
  }

  for(y = 0; y < img->height; y++)
    for(x = 0; x < 3*img->width; x+=3)
    {
      r = data[y][x];
      g = data[y][x+1];
      b = data[y][x+2];
      Y = (6969*r + 23434*g + 2365*b)/32768;
      data[y][x]   = (R[Y] * density + r * (100-density)) / 100;
      data[y][x+1] = (G[Y] * density + g * (100-density)) / 100;
      data[y][x+2] = (B[Y] * density + b * (100-density)) / 100;
    }
}


